#ifndef natural_h
#define natural_h

#include <assert.h>
#include <algorithm>
#include <vector>
#include <gmpxx.h>
#include "chain.h"
#include "bit.h"

#define natural_opsnum 4

class natural {
  std::vector<bit> n; /* little-endian */

  static long long nummult;
  static long long numsq;
  static long long nummod;
  static long long numinvert;
public:
  static long long ops(int t) {
    if (t == 1) return numsq;
    if (t == 2) return nummod;
    if (t == 3) return numinvert;
    return nummult;
  }
  static const char *opsname(int t) {
    if (t == 1) return "sq";
    if (t == 2) return "mod";
    if (t == 3) return "invert";
    return "mult";
  }

  natural() {
  }
  natural(bit b) {
    n.push_back(b);
  }
  natural(std::vector< std::vector<bit> > x,long long outlen) {
    // semantics: sum(2^i * sum(x[i])) mod 2^outlen
    x.resize(outlen);
    for (long long i = 0;i < outlen;++i) {
      while (x[i].size() >= 3) {
        bit a = x[i].back(); x[i].pop_back();
        bit b = x[i].back(); x[i].pop_back();
        bit ab = a ^ b;
        bit c = x[i].back(); x[i].pop_back();
        x[i].push_back(ab ^ c);
        if (i+1 < outlen)
          x[i+1].push_back((a & b) ^ (ab & c));
      }
      if (x[i].size() >= 2) {
        bit a = x[i].back(); x[i].pop_back();
        bit b = x[i].back(); x[i].pop_back();
        x[i].push_back(a ^ b);
        if (i+1 < outlen)
          x[i+1].push_back(a & b);
      }
      if (x[i].size() == 1) {
        n.push_back(x[i][0]);
      } else {
        n.push_back(bit(0));
      }
    }
  }
  natural(mpz_class m,long long bits) {
    assert(m >= 0);
    n.reserve(bits);
    while (bits > 0) {
      mpz_class mbit = m & 1;
      n.push_back(mbit.get_si());
      m >>= 1;
      bits -= 1;
    }
  }
  natural(mpz_class m) {
    if (m == 0) *this = natural();
    else *this = natural(m,mpz_sizeinbase(m.get_mpz_t(),2));
  }
  natural(int n) {
    *this = natural(mpz_class(n));
  }

  mpz_class value() const {
    mpz_class result = 0;
    for (long long i = n.size() - 1;i >= 0;--i)
      result = (result << 1) + n[i].value();
    return result;
  }

  long long bits() const {
    return n.size();
  }
  bit bitatpos(long long i) const {
    if (i < 0) return 0;
    if (i >= n.size()) return 0;
    return n[i];
  }
  
  bit isequal(mpz_class m) const {
    long long nlen = n.size();
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);

    if (mlen > nlen) {
      if (m == 0 && nlen == 0) return bit(1);
      return bit(0);
    }

    if (nlen == 0) return bit(1);

    bit result;
    if (0 < mlen && mpz_tstbit(m.get_mpz_t(),0))
      result = n[0];
    else
      result = ~n[0];
    for (long long i = 1;i < nlen;++i)
      if (i < mlen && mpz_tstbit(m.get_mpz_t(),i))
        result &= n[i];
      else
        result &= ~n[i];

    return result;
  }

  natural operator^(const bit &flip) const {
    long long nlen = n.size();
    if (nlen == 0) return natural(flip);
    natural result;
    result.n.push_back(n[0] ^ flip);
    for (long long i = 1;i < nlen;++i)
      result.n.push_back(n[i]);
    return result;
  }

  // return m if control, otherwise this
  natural mux(const natural &m,const bit &control) const {
    long long nlen = n.size();
    long long mlen = m.n.size();
    long long i;
    natural result;
    for (i = 0;i < mlen && i < nlen;++i)
      result.n.push_back(n[i] ^ (control & (n[i] ^ m.n[i])));
    if (i < nlen) {
      bit notcontrol = ~control;
      for (;i < nlen;++i)
        result.n.push_back(n[i] & notcontrol);
    }
    for (;i < mlen;++i)
      result.n.push_back(m.n[i] & control);
    return result;
  }

  void cswap(natural &m,const bit &control) {
    long long nlen = n.size();
    long long mlen = m.n.size();
    long long i;
    for (i = 0;i < mlen && i < nlen;++i) {
      bit flip = control & (n[i] ^ m.n[i]);
      n[i] ^= flip;
      m.n[i] ^= flip;
    }
    for (;i < nlen;++i) {
      bit flip = control & n[i];
      n[i] ^= flip;
      m.n.push_back(flip);
    }
    for (;i < mlen;++i) {
      bit flip = control & m.n[i];
      n.push_back(flip);
      m.n[i] ^= flip;
    }
  }

  natural operator>>(long long shift) const {
    natural result;
    long long i;
    for (i = shift;i < 0;++i)
      result.n.push_back(0);
    for (;i < n.size();++i)
      result.n.push_back(n[i]);
    return result;
  }
  natural bottom(long long shift) const {
    natural result;
    for (long long i = 0;i < n.size() && i < shift;++i)
      result.n.push_back(n[i]);
    return result;
  }

  natural operator<<(long long shift) const {
    natural result;
    for (long long i = 0;i < shift;++i)
      result.n.push_back(bit(0));
    for (long long i = 0;i < n.size();++i)
      result.n.push_back(n[i]);
    return result;
  }
  natural dbl() const {
    return *this << 1;
  }

  natural operator+(const natural &m) const {
    natural result;
    long long nlen = n.size();
    long long mlen = m.n.size();
    long long i;
    if (mlen == 0) return *this;
    if (nlen == 0) return m;
    result.n.push_back(m.n[0] ^ n[0]);
    bit carry = m.n[0] & n[0];
    for (i = 1;i < mlen && i < nlen;++i) {
      bit x0 = m.n[i] ^ n[i];
      bit x1 = m.n[i] & n[i];
      x1 ^= x0 & carry;
      x0 ^= carry;
      result.n.push_back(x0);
      carry = x1;
    }
    for (;i < mlen;++i) {
      result.n.push_back(m.n[i] ^ carry);
      carry &= m.n[i];
    }
    for (;i < nlen;++i) {
      result.n.push_back(n[i] ^ carry);
      carry &= n[i];
    }
    result.n.push_back(carry);
    return result;
  }
  natural operator+(const mpz_class &m) const {
    if (m == 0) return *this;
    natural result;
    long long nlen = n.size();
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);
    long long i;
    if (nlen == 0) return natural(m);
    bit carry;
    if (mpz_tstbit(m.get_mpz_t(),0)) {
      result.n.push_back(~n[0]);
      carry = n[0];
    } else {
      result.n.push_back(n[0]);
      carry = 0;
    }
    for (i = 1;i < mlen && i < nlen;++i) {
      bit x0;
      bit x1;
      if (mpz_tstbit(m.get_mpz_t(),i)) {
        x0 = ~(n[i] ^ carry);
        x1 = n[i] | carry;
      } else {
        x0 = n[i] ^ carry;
        x1 = n[i] & carry;
      }
      result.n.push_back(x0);
      carry = x1;
    }
    for (;i < mlen;++i) {
      if (mpz_tstbit(m.get_mpz_t(),i)) {
        result.n.push_back(~carry);
      } else {
        result.n.push_back(carry);
        carry = 0;
      }
    }
    for (;i < nlen;++i) {
      result.n.push_back(n[i] ^ carry);
      carry &= n[i];
    }
    result.n.push_back(carry);
    return result;
  }
  natural operator+(const int &m) const {
    return *this + mpz_class(m);
  }

  natural subtract(const natural &m,long long maxlen) const {
    natural result;
    long long nlen = n.size();
    long long mlen = m.n.size();
    long long i;
    if (mlen == 0) return *this;
    bit carry = 0;
    for (i = 0;i < mlen && i < nlen && i < maxlen;++i) {
      bit x0 = m.n[i] ^ n[i];
      if (i + 1 < maxlen) {
        bit x1 = m.n[i] & ~n[i];
        x1 ^= ~x0 & carry;
        x0 ^= carry;
        result.n.push_back(x0);
        carry = x1;
      } else {
        x0 ^= carry;
        result.n.push_back(x0);
      }
    }
    for (;i < mlen && i < maxlen;++i) {
      result.n.push_back(m.n[i] ^ carry);
      if (i + 1 < maxlen) carry |= m.n[i];
    }
    for (;i < nlen && i < maxlen;++i) {
      result.n.push_back(n[i] ^ carry);
      if (i + 1 < maxlen) carry &= ~n[i];
    }
    for (;i < maxlen;++i)
      result.n.push_back(carry);
    return result;
  }
  natural subtract(const mpz_class &m,long long maxlen) const {
    if (m == 0) return *this;
    natural result;
    long long nlen = n.size();
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);
    long long i;
    bit carry = 0;
    for (i = 0;i < mlen && i < nlen && i < maxlen;++i) {
      bit x0,x1;
      if (mpz_tstbit(m.get_mpz_t(),i)) {
        x0 = ~n[i];
        x1 = x0 | carry;
        x0 ^= carry;
      } else {
        x0 = n[i];
        x1 = ~x0 & carry;
        x0 ^= carry;
      }
      result.n.push_back(x0);
      carry = x1;
    }
    for (;i < mlen && i < maxlen;++i) {
      if (mpz_tstbit(m.get_mpz_t(),i)) {
        result.n.push_back(~carry);
        carry = 1;
      } else {
        result.n.push_back(carry);
      }
    }
    for (;i < nlen && i < maxlen;++i) {
      result.n.push_back(n[i] ^ carry);
      carry &= ~n[i];
    }
    for (;i < maxlen;++i)
      result.n.push_back(carry);
    return result;
  }

  natural times(const natural &m) const {
    long long nlen = n.size();
    long long mlen = m.n.size();
    if (mlen == 0) return m;
    if (nlen == 0) return *this;
    long long outlen = mlen + nlen;
    if (mlen == 1) outlen = nlen;
    if (nlen == 1) outlen = mlen;

    if (std::min(mlen,nlen) == 14 || std::min(mlen,nlen) >= 16) {
      long long mid = std::min(mlen/2,nlen/2);

      natural m0 = m.bottom(mid); natural m1 = m >> mid;
      natural n0 = (*this).bottom(mid); natural n1 = *this >> mid;

      natural h0 = m0.times(n0);
      natural h1 = (m0+m1).times(n0+n1);
      natural h2 = m1.times(n1);

      h1 = h1.subtract(h0+h2,std::max(mlen,nlen)+1);

      std::vector< std::vector<bit> > x;
      x.resize(outlen);
      for (long long i = 0;i < h0.n.size();++i)
        if (i < outlen)
          x[i].push_back(h0.n[i]);
      for (long long i = 0;i < h1.n.size();++i)
        if (i+mid < outlen)
          x[i+mid].push_back(h1.n[i]);
      for (long long i = 0;i < h2.n.size();++i)
        if (i+2*mid < outlen)
          x[i+2*mid].push_back(h2.n[i]);

      return natural(x,outlen);
    }

    std::vector< std::vector<bit> > x;
    x.resize(outlen);
    for (long long i = 0;i < mlen;++i)
      for (long long j = 0;j < nlen;++j)
        x[i+j].push_back(m.n[i] & n[j]);
    return natural(x,outlen);
  }
  natural times(const mpz_class &m,long long trunc = 9223372036854775807) const {
    if (trunc < 0) trunc = 0;
    if (trunc == 0) return natural(0);

    assert(m >= 0);
    if (m == 0) return natural(0);
    long long nlen = n.size();
    if (nlen == 0) return natural(0);
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);
    if (nlen > trunc) nlen = trunc;
    if (mlen > trunc) mlen = trunc;

    long long outlen = mlen + nlen;
    if (mlen == 1) outlen = nlen;
    if (nlen == 1) outlen = mlen;
    if (outlen > trunc) outlen = trunc;

    if (mlen >= 30 && nlen >= 30) {
      long long mid = std::min(mlen/2,nlen/2);

      mpz_class m0 = m & ((mpz_class(1) << mid) - 1); mpz_class m1 = m >> mid;
      natural n0 = (*this).bottom(mid); natural n1 = *this >> mid;

      natural h0 = n0.times(m0,outlen);
      natural h1 = (n0+n1).times(m0+m1,outlen-mid);
      natural h2 = n1.times(m1,outlen-mid);

      h1 = h1.subtract(h0+h2,std::max(mlen,nlen)+1);

      std::vector< std::vector<bit> > x;
      x.resize(outlen);
      for (long long i = 0;i < h0.n.size();++i)
        if (i < outlen)
          x[i].push_back(h0.n[i]);
      for (long long i = 0;i < h1.n.size();++i)
        if (i+mid < outlen)
          x[i+mid].push_back(h1.n[i]);
      for (long long i = 0;i < h2.n.size();++i)
        if (i+2*mid < outlen)
          x[i+2*mid].push_back(h2.n[i]);

      return natural(x,outlen);
    }

    std::vector< std::vector<bit> > x;
    x.resize(outlen);
    for (long long i = 0;i < mlen;++i)
      if (mpz_tstbit(m.get_mpz_t(),i))
        for (long long j = 0;j < nlen;++j)
          if (i+j < outlen)
            x[i+j].push_back(n[j]);
    return natural(x,outlen);
  }
  natural timesself() const {
    long long nlen = n.size();
    if (nlen == 0) return *this;
    long long outlen = nlen + nlen;
    if (nlen == 1) outlen = nlen;

    if (nlen >= 22) {
      long long mid = nlen/2;

      natural n0 = (*this).bottom(mid); natural n1 = *this >> mid;

      natural h0 = n0.timesself();
      natural h1 = (n0+n1).timesself();
      natural h2 = n1.timesself();

      h1 = h1.subtract(h0+h2,nlen+1);

      std::vector< std::vector<bit> > x;
      x.resize(outlen);
      for (long long i = 0;i < h0.n.size();++i)
        if (i < outlen)
          x[i].push_back(h0.n[i]);
      for (long long i = 0;i < h1.n.size();++i)
        if (i+mid < outlen)
          x[i+mid].push_back(h1.n[i]);
      for (long long i = 0;i < h2.n.size();++i)
        if (i+2*mid < outlen)
          x[i+2*mid].push_back(h2.n[i]);

      return natural(x,outlen);
    }

    std::vector< std::vector<bit> > x;
    x.resize(outlen);
    for (long long i = 0;i < nlen;++i)
      x[i+i].push_back(n[i]);
    for (long long i = 0;i < nlen;++i)
      for (long long j = 0;j < i;++j)
        x[i+j+1].push_back(n[i] & n[j]);

    return natural(x,outlen);
  }

  natural operator*(const natural &m) const {
    ++nummult;
    return times(m);
  }
  natural operator*(const mpz_class &m) const {
    ++nummult;
    return times(m);
  }
  natural square() {
    ++numsq;
    return (*this).timesself();
  }

  // input required to be between 0 and 2m-1
  natural minimod(mpz_class m) const {
    long long nlen = n.size();
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);

    if (nlen < mlen) return *this;

    natural diff = (*this).subtract(m,mlen + 1);
    bit select = diff.n[mlen];

    natural result;
    for (long long i = 0;i < mlen;++i)
      result.n.push_back(diff.n[i] ^ (select & (diff.n[i] ^ n[i])));
    return result;
  }

  natural mod(mpz_class m) const {
    if (m < 0) m = -m;
    if (m == 0) return *this;
    if (m == 1) return natural(0);
    long long nlen = n.size();
    long long mlen = mpz_sizeinbase(m.get_mpz_t(),2);

    if (nlen < mlen) return *this;

    mpz_class precomp = (mpz_class(1) << (nlen + 2)) / m;
    natural q = (*this >> (mlen - 2)).times(precomp) >> (nlen + 4 - mlen);
    natural result = (*this).subtract(q.times(m,mlen + 1),mlen + 1);
    return result.minimod(m);
  }
  natural operator%(mpz_class m) const {
    ++nummod;
    return (*this).mod(m);
  }
  natural operator%=(mpz_class m) { *this = *this % m; return *this; }

  natural subtractmod(const natural &m,mpz_class p) const {
    long long plen = mpz_sizeinbase(p.get_mpz_t(),2);
    natural x = (*this).mod(p);
    natural y = m.mod(p);
    natural xy = (x + p).subtract(y,plen + 1);
    return xy.minimod(p);
  }

  // assumes input is already reduced mod p
  natural halfmod(mpz_class p) const {
    bit odd = (*this).bitatpos(0);
    natural x = natural(0).mux(p,odd) + *this;
    return x>>1;
  }

  // return eth power of *this mod m
  natural powmod(const natural e,mpz_class m) const {
    if (e.bits() == 0)
      return natural(1) % m;

    natural result;
    bool known1 = true;

    if (e.bits() >= 4) {
      natural precomp[4];
      precomp[0] = natural(1);
      precomp[1] = *this;
      precomp[2] = precomp[1].square() % m;
      precomp[3] = (precomp[1] * precomp[2]) % m;
      for (long long i = (e.bits() - 1) & ~1;i >= 0;i -= 2) {
        if (!known1) {
          result = result.square() % m;
          result = result.square() % m;
        }
        natural ram[4];
        ram[0] = precomp[0];
        ram[1] = precomp[1];
        ram[2] = precomp[2];
        ram[3] = precomp[3];
        for (long long j = i + 1;j >= i;--j) {
          if (j >= e.bits()) continue;
          if (j == i + 1) {
            ram[0] = ram[0].mux(ram[2],e.bitatpos(j));
            ram[1] = ram[1].mux(ram[3],e.bitatpos(j));
          } else {
            ram[0] = ram[0].mux(ram[1],e.bitatpos(j));
          }
        }
        if (known1)
          result = ram[0] % m;
        else
          result = (result * ram[0]) % m;
        known1 = false;
      }
      return result;
    }

    for (long long i = e.bits() - 1;i >= 0;--i) {
      if (!known1)
        result = result.square() % m;
      natural factor(1);
      factor = factor.mux(*this,e.bitatpos(i));
      if (known1)
        result = factor % m;
      else
        result = (result * factor) % m;
      known1 = false;
    }
    return result;
  }

  // return (*this)^e x^8 mod m
  natural powpow8mod(const natural e,natural x,mpz_class m) const {
    long long bits = e.bits();
    if (bits < 5) bits = 5;

    natural result;
    bool known1 = true;

    natural precomp[4];
    precomp[0] = natural(1);
    precomp[1] = *this;
    precomp[2] = precomp[1].square() % m;
    precomp[3] = (precomp[1] * precomp[2]) % m;
    for (long long i = (bits - 1) & ~1;i >= 0;i -= 2) {
      if (!known1) {
        result = result.square() % m;
        if (i == 2) result = (result*x) % m;
        result = result.square() % m;
      }
      natural ram[4];
      ram[0] = precomp[0];
      ram[1] = precomp[1];
      ram[2] = precomp[2];
      ram[3] = precomp[3];
      for (long long j = i + 1;j >= i;--j) {
        if (j >= e.bits()) continue;
        if (j == i + 1) {
          ram[0] = ram[0].mux(ram[2],e.bitatpos(j));
          ram[1] = ram[1].mux(ram[3],e.bitatpos(j));
        } else {
          ram[0] = ram[0].mux(ram[1],e.bitatpos(j));
        }
      }
      if (known1)
        result = ram[0] % m;
      else
        result = (result * ram[0]) % m;
      known1 = false;
    }
    return result;
  }

  // nonzero W is required to be chain_bestW(e)
  natural powmod(const mpz_class &e,mpz_class m,int W = 0) const {
    if (e == 0) return natural(1) % m;
    assert(e > 0);
    if (e == 1) return *this % m;

    if (!W) W = chain_bestW(e);
    std::vector<int> ops = chain(e,W);

    std::vector<natural> precomp;
    precomp.resize(W + 1);
    precomp[1] = *this;
    if (W >= 3) {
      precomp[2] = precomp[1].square() % m;
      for (int e = 3;e <= W;e += 2)
        precomp[e] = (precomp[2] * precomp[e - 2]) % m;
    }

    natural result = precomp[ops.back()];
    ops.pop_back();

    while (ops.size() > 0) {
      int r = ops.back();
      ops.pop_back();

      if (r == 0)
        result = result.square() % m;
      else
        result = (precomp[r] * result) % m;
    }
    return result;
  }

  // return reciprocal of input mod p
  // exception: for input 0 return 0
  // exception: for input 0 return 1 if p is 2
  // nonzero W is required to be chain_bestW(p-2)
  natural invertmodprime(const mpz_class &p,int W = 0) const {
    ++numinvert;
    return (*this).powmod(p-2,p,W);
  }
} ;

#endif
